package dev.fyloz.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.config.security.defaultGroupCookieName
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.account.*
import dev.fyloz.colorrecipesexplorer.repository.UserRepository
import dev.fyloz.colorrecipesexplorer.repository.GroupRepository
import org.junit.jupiter.api.*
import org.springframework.mock.web.MockHttpServletResponse
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import java.util.*
import javax.servlet.http.Cookie
import javax.servlet.http.HttpServletRequest
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertNotNull
import kotlin.test.assertTrue
import org.springframework.security.core.userdetails.User as SpringUser

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class UserServiceTest :
  AbstractExternalModelServiceTest<User, UserSaveDto, UserUpdateDto, UserService, UserRepository>() {
    private val passwordEncoder = BCryptPasswordEncoder()

    override val entity: User = user(passwordEncoder, id = 0L)
    override val anotherEntity: User = user(passwordEncoder, id = 1L)
    private val entityDefaultGroupUser = user(passwordEncoder, id = 2L, isDefaultGroupUser = true)
    private val entitySystemUser = user(passwordEncoder, id = 3L, isSystemUser = true)
    private val group = group(id = 0L)
    override val entitySaveDto: UserSaveDto = spy(userSaveDto(passwordEncoder, id = 0L))
    override val entityUpdateDto: UserUpdateDto = spy(userUpdateDto(id = 0L))

    override val repository: UserRepository = mock()
    private val groupService: GroupService = mock()
    override val service: UserService = spy(UserServiceImpl(repository, groupService, passwordEncoder))

    private val entitySaveDtoUser = User(
      entitySaveDto.id,
      entitySaveDto.firstName,
      entitySaveDto.lastName,
      passwordEncoder.encode(entitySaveDto.password),
      isDefaultGroupUser = false,
      isSystemUser = false,
      group = null,
      permissions = entitySaveDto.permissions
    )

    @AfterEach
    override fun afterEach() {
        reset(groupService)
        super.afterEach()
    }

    // existsByFirstNameAndLastName()

    @Test
    fun `existsByFirstNameAndLastName() returns true when an user with the given first name and last name exists`() {
        whenever(repository.existsByFirstNameAndLastName(entity.firstName, entity.lastName)).doReturn(true)

        val found = service.existsByFirstNameAndLastName(entity.firstName, entity.lastName)

        assertTrue(found)
    }

    @Test
    fun `existsByFirstNameAndLastName() returns false when no user with the given first name and last name exists`() {
        whenever(repository.existsByFirstNameAndLastName(entity.firstName, entity.lastName)).doReturn(false)

        val found = service.existsByFirstNameAndLastName(entity.firstName, entity.lastName)

        assertFalse(found)
    }

    // getById()

    @Test
    fun `getById() throws NotFoundException when the corresponding user is a default group user`() {
        whenever(repository.findById(entityDefaultGroupUser.id)).doReturn(Optional.of(entityDefaultGroupUser))

        assertThrows<NotFoundException> {
            service.getById(
              entityDefaultGroupUser.id,
              ignoreDefaultGroupUsers = true,
              ignoreSystemUsers = false
            )
        }.assertErrorCode()
    }

    @Test
    fun `getById() throws NotFoundException when the corresponding user is a system user`() {
        whenever(repository.findById(entitySystemUser.id)).doReturn(Optional.of(entitySystemUser))

        assertThrows<NotFoundException> {
            service.getById(
              entitySystemUser.id,
              ignoreDefaultGroupUsers = false,
              ignoreSystemUsers = true
            )
        }.assertErrorCode()
    }

    // getByGroup()

    @Test
    fun `getByGroup() returns all the users with the given group from the repository`() {
        whenever(repository.findAllByGroup(group)).doReturn(entityList)

        val found = service.getByGroup(group)

        assertTrue(found.containsAll(entityList))
        assertTrue(entityList.containsAll(found))
    }

    @Test
    fun `getByGroup() returns an empty list when there is no user with the given group in the repository`() {
        whenever(repository.findAllByGroup(group)).doReturn(listOf())

        val found = service.getByGroup(group)

        assertTrue(found.isEmpty())
    }

    // getDefaultGroupUser()

    @Test
    fun `getDefaultGroupUser() returns the default user of the given group from the repository`() {
        whenever(repository.findByIsDefaultGroupUserIsTrueAndGroupIs(group)).doReturn(entityDefaultGroupUser)

        val found = service.getDefaultGroupUser(group)

        assertEquals(entityDefaultGroupUser, found)
    }

    // save()

    override fun `save() saves in the repository and returns the saved value`() {
        whenever(repository.save(entity)).doReturn(entity)
        doReturn(false).whenever(repository).existsByFirstNameAndLastName(entity.firstName, entity.lastName)

        val found = service.save(entity)

        verify(repository).save(entity)
        assertEquals(entity, found)
    }

    @Test
    fun `save() throws AlreadyExistsException when firstName and lastName exists`() {
        doReturn(true).whenever(repository).existsByFirstNameAndLastName(entity.firstName, entity.lastName)

        assertThrows<AlreadyExistsException> { service.save(entity) }
          .assertErrorCode("fullName")
    }

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        withBaseSaveDtoTest(entity, entitySaveDto, service, {
            argThat {
                this.id == entity.id && this.firstName == entity.firstName && this.lastName == entity.lastName
            }
        })
    }

    @Test
    fun `save(dto) calls and returns save() with the created user`() {
        doReturn(entitySaveDtoUser).whenever(service).save(any<User>())

        val found = service.save(entitySaveDto)

        verify(service).save(argThat<User> { this.id == entity.id && this.firstName == entity.firstName && this.lastName == entity.lastName })
        assertEquals(entitySaveDtoUser, found)
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() =
      withBaseUpdateDtoTest(entity, entityUpdateDto, service, { any() })

    @Test
    fun `update() throws AlreadyExistsException when a different user with the given first name and last name exists`() {
        whenever(repository.findByFirstNameAndLastName(entity.firstName, entity.lastName)).doReturn(
          entityDefaultGroupUser
        )
        doReturn(entity).whenever(service).getById(eq(entity.id), any(), any())

        assertThrows<AlreadyExistsException> {
            service.update(
              entity,
              true,
              ignoreSystemUsers = true
            )
        }.assertErrorCode("fullName")
    }
}

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class GroupServiceTest :
  AbstractExternalNamedModelServiceTest<Group, GroupSaveDto, GroupUpdateDto, GroupService, GroupRepository>() {
    private val userService: UserService = mock()
    override val repository: GroupRepository = mock()
    override val service: GroupServiceImpl = spy(GroupServiceImpl(userService, repository))

    override val entity: Group = group(id = 0L, name = "group")
    override val anotherEntity: Group = group(id = 1L, name = "another group")
    override val entitySaveDto: GroupSaveDto = spy(groupSaveDto(name = "group"))
    override val entityUpdateDto: GroupUpdateDto = spy(groupUpdateDto(id = 0L, name = "group"))
    override val entityWithEntityName: Group = group(id = 2L, name = entity.name)

    private val groupUserId = 1000000L + entity.id!!
    private val groupUser = user(BCryptPasswordEncoder(), id = groupUserId, group = entity)

    @BeforeEach
    override fun afterEach() {
        reset(userService)
        super.afterEach()
    }

    // getUsersForGroup()

    @Test
    fun `getUsersForGroup() returns all users in the given group`() {
        val group = group(id = 1L)

        doReturn(group).whenever(service).getById(group.id!!)
        whenever(userService.getByGroup(group)).doReturn(listOf(groupUser))

        val found = service.getUsersForGroup(group.id!!)

        assertTrue(found.contains(groupUser))
        assertTrue(found.size == 1)
    }

    @Test
    fun `getUsersForGroup() returns empty collection when the given group contains any user`() {
        doReturn(entity).whenever(service).getById(entity.id!!)

        val found = service.getUsersForGroup(entity.id!!)

        assertTrue(found.isEmpty())
    }

    // getRequestDefaultGroup()

    @Test
    fun `getRequestDefaultGroup() returns the group contained in the cookie of the HTTP request`() {
        val cookies: Array<Cookie> = arrayOf(Cookie(defaultGroupCookieName, groupUserId.toString()))
        val request: HttpServletRequest = mock()

        whenever(request.cookies).doReturn(cookies)
        whenever(userService.getById(eq(groupUserId), any(), any())).doReturn(groupUser)

        val found = service.getRequestDefaultGroup(request)

        assertEquals(entity, found)
    }

    @Test
    fun `getRequestDefaultGroup() throws NoDefaultGroupException when the HTTP request does not contains a cookie for the default group`() {
        val request: HttpServletRequest = mock()

        whenever(request.cookies).doReturn(arrayOf())

        assertThrows<NoDefaultGroupException> { service.getRequestDefaultGroup(request) }
    }

    // setResponseDefaultGroup()

    @Test
    fun `setResponseDefaultGroup() the default group cookie has been added to the given HTTP response with the given group id`() {
        val response = MockHttpServletResponse()

        whenever(userService.getDefaultGroupUser(entity)).doReturn(groupUser)
        doReturn(entity).whenever(service).getById(entity.id!!)

        service.setResponseDefaultGroup(entity.id!!, response)
        val found = response.getCookie(defaultGroupCookieName)

        assertNotNull(found)
        assertEquals(defaultGroupCookieName, found.name)
        assertEquals(groupUserId.toString(), found.value)
        assertEquals(defaultGroupCookieMaxAge, found.maxAge)
        assertTrue(found.isHttpOnly)
        assertTrue(found.secure)
    }

    // save()

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        withBaseSaveDtoTest(entity, entitySaveDto, service)
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() =
      withBaseUpdateDtoTest(entity, entityUpdateDto, service, { any() })
}

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class UserUserDetailsServiceTest {
    private val userService: UserService = mock()
    private val service = spy(CreUserDetailsServiceImpl(userService))

    private val user = user(id = 0L)

    @BeforeEach
    fun beforeEach() {
        reset(userService, service)
    }

    // loadUserByUsername()

    @Test
    fun `loadUserByUsername() calls loadUserByUserId() with the given username as an id`() {
        whenever(userService.getById(eq(user.id), any(), any())).doReturn(user)
        doReturn(SpringUser(user.id.toString(), user.password, listOf())).whenever(service)
          .loadUserById(user.id)

        service.loadUserByUsername(user.id.toString())

        verify(service).loadUserById(eq(user.id), any())
    }

    @Test
    fun `loadUserByUsername() throws UsernameNotFoundException when no user with the given id exists`() {
        whenever(userService.getById(eq(user.id), any(), any())).doThrow(
          userIdNotFoundException(user.id)
        )

        assertThrows<UsernameNotFoundException> { service.loadUserByUsername(user.id.toString()) }
    }

    // loadUserByUserId

    @Test
    fun `loadUserByUserId() returns an User corresponding to the user with the given id`() {
        whenever(userService.getById(eq(user.id), any(), any())).doReturn(user)

        val found = service.loadUserById(user.id)

        assertEquals(user.id, found.username.toLong())
        assertEquals(user.password, found.password)
    }
}
